from __future__ import annotations

import json
import math
from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar

import torch
import torch.nn as nn
import torch.nn.functional as F

from babybert.utils import resolve_checkpoint_path


@dataclass
class BabyBERTConfig:
    """
    Configuration class for BabyBERT.

    Attributes:
        vocab_size: The vocab size of the tokenizer used to create inputs.
        block_size: The length of the input sequences.
        hidden_size: The size of the model's hidden states. Must be divisible by
                     `n_heads`.
        n_blocks: The number of transformer blocks to use.
        n_heads: The number of attention heads to use.
        attention_dropout_probability: The dropout probability for the output of each
                                       attention head.
        attention_projection_dropout_probability: The dropout probability for the
                                                  output projection of a multi-head
                                                  self-attention block.
        mlp_dropout_probability: The dropout probability for the MLP used for tokenwise
                                 learning.
        embedding_dropout_probability: The dropout probability for input sequence
                                       embeddings.
    """

    vocab_size: int
    block_size: int
    hidden_size: int = 768
    n_blocks: int = 12
    n_heads: int = 12
    attention_dropout_probability: float = 0.1
    attention_projection_dropout_probability: float = 0.1
    mlp_dropout_probability: float = 0.1
    embedding_dropout_probability: float = 0.1
    ignore_index: int = -100

    _PRESETS: ClassVar[dict[str, dict]] = {
        "tiny": {
            "hidden_size": 128,
            "n_blocks": 2,
            "n_heads": 2,
        },
        "small": {
            "hidden_size": 256,
            "n_blocks": 4,
            "n_heads": 4,
        },
        "base": {
            "hidden_size": 768,
            "n_blocks": 12,
            "n_heads": 12,
        },
    }

    @classmethod
    def from_preset(cls, name: str, **kwargs) -> BabyBERTConfig:
        """
        Load a configuration object from a preset.

        **Note**: The configuration presets only contain values for `hidden_size`,
        `n_blocks`, and `n_heads`. Other required attributes without defaults, such as
        `vocab_size` and `block_size`, must be passed in as keyword arguments to this
        method. Keyword arguments with defaults can optionally be included as well.

        **Available presets**:
        - `"tiny"`: `hidden_size=128`, `n_blocks=2`, `n_heads=2`
        - `"small"`: `hidden_size=256`, `n_blocks=4`, `n_heads=4`
        - `"base"`: `hidden_size=768`, `n_blocks=12`, `n_heads=12`

        Args:
            name: The name of the configuration preset from which to load.
        Returns:
            A new configuration object, with values populated by the defaults.
        """
        if name not in cls._PRESETS:
            available = ", ".join(f"'{key}'" for key in cls._PRESETS.keys())
            raise ValueError(
                f"Preset config '{name}' is not available."
                f"Available presets: {available}"
            )
        return cls(**cls._PRESETS[name], **kwargs)


class AttentionHead(nn.Module):
    """Implementation of a single self-attention head."""

    def __init__(self, config: BabyBERTConfig):
        super().__init__()

        # The size of each attention head's latent space is equal to the embedding
        # space's size divided by the number of heads. That way, after we compute the
        # context vector for each attention head, we can concatenate them to produce a
        # vector that's the same size as the original embedding space.
        head_size = config.hidden_size // config.n_heads

        # Layers for projecting the input embeddings into three spaces:
        # the query (Q) space, the key (K) space, and the value (V) space.
        self.query_embedding = nn.Linear(config.hidden_size, head_size)
        self.key_embedding = nn.Linear(config.hidden_size, head_size)
        self.value_embedding = nn.Linear(config.hidden_size, head_size)

        self.dropout = nn.Dropout(config.attention_dropout_probability)

    def forward(self, x, mask=None):
        # Project the inputs into each space.
        q = self.query_embedding(x)
        k = self.key_embedding(x)
        v = self.value_embedding(x)

        # Perform scaled dot product attention (SDPA) using the query and key matrices
        # - this yields the similarity between each embedding in our input sequences.
        raw_attention_weights = (q @ torch.transpose(k, -2, -1)) / math.sqrt(q.size(-1))

        # If we have an attention mask, we replace each of the masked tokens with a
        # negative infinity in our attention weight matrix. This makes sure that the
        # attention score will be 0 when we compute the softmax later.
        if mask is not None:
            mask = mask.unsqueeze(1)
            raw_attention_weights = raw_attention_weights.masked_fill(
                mask == 0, float("-inf")
            )

        # Compute the softmax over each input sequence, normalizing the scores.
        normalized_attention_weights = F.softmax(raw_attention_weights, -1)
        normalized_attention_weights = self.dropout(normalized_attention_weights)

        # Taking the dot product of the weight matrix and the value matrix gives us our
        # final context vector.
        context_vector = normalized_attention_weights @ v

        return context_vector


class MultiHeadSelfAttention(nn.Module):
    """Implementation of a multi-head self-attention layer."""

    def __init__(self, config: BabyBERTConfig):
        super().__init__()

        # Initialize n attention heads - this allows us to capture n sets of
        # relationships between token embeddings.
        self.attention_heads = nn.ModuleList(
            [AttentionHead(config) for _ in range(config.n_heads)]
        )
        self.output_projection = nn.Linear(config.hidden_size, config.hidden_size)
        self.dropout = nn.Dropout(config.attention_projection_dropout_probability)

    def forward(self, x, mask=None):
        # Concatenate the output of each of the n attention heads.
        x = torch.cat(
            [attention_head(x, mask) for attention_head in self.attention_heads], dim=-1
        )

        # Project the concatenated vectors to the output space.
        x = self.output_projection(x)
        x = self.dropout(x)
        return x


class TransformerBlock(nn.Module):
    """Implementation of a transformer block."""

    def __init__(self, config: BabyBERTConfig):
        super().__init__()

        self.mhsa = MultiHeadSelfAttention(config)
        self.layer_norm_1 = nn.LayerNorm(config.hidden_size)

        # Simple multilayer perceptron (MLP) to handle tokenwise learning.
        self.mlp = nn.Sequential(
            nn.Linear(config.hidden_size, config.hidden_size * 4),
            nn.GELU(),
            nn.Linear(config.hidden_size * 4, config.hidden_size),
            nn.Dropout(config.mlp_dropout_probability),
        )
        self.layer_norm_2 = nn.LayerNorm(config.hidden_size)

    def forward(self, x, mask=None):
        # Pass the input through our multi-head self attention layer.
        x = x + self.mhsa(x, mask)
        x = self.layer_norm_1(x)

        # Pass the input through the MLP.
        x = x + self.mlp(x)
        x = self.layer_norm_2(x)
        return x


class BabyBERT(nn.Module):
    """Minimal implementation of [BERT](https://arxiv.org/pdf/1810.04805)."""

    def __init__(self, config: BabyBERTConfig):
        super().__init__()

        self.config = config

        # Learned embeddings for each token ID in the vocabulary.
        self.token_embeddings = nn.Embedding(config.vocab_size, config.hidden_size)

        # Learned embeddings for each segment; we only have segments 1 and 2 in BERT.
        self.segment_embeddings = nn.Embedding(2, config.hidden_size)

        # Learned embeddings for each token position.
        # Having positional embeddings is necessary for the model to understand the
        # order of the input tokens.
        self.positional_embeddings = nn.Embedding(config.block_size, config.hidden_size)

        self.dropout = nn.Dropout(config.embedding_dropout_probability)
        self.transformer_blocks = nn.ModuleList(
            [TransformerBlock(config) for _ in range(config.n_blocks)]
        )

    def forward(self, token_ids, attention_mask=None, segment_encodings=None):
        # Get the position IDs - these range from 0 to the max length of an input
        # sequence (stored in the block_size variable).
        position_ids = torch.arange(
            0, self.config.block_size, device=token_ids.device
        ).unsqueeze(0)

        # If no segment encodings tensor was passed in, assume everything belongs
        # to the first segment (represented by '0')
        if segment_encodings is None:
            segment_encodings = torch.zeros_like(token_ids)

        # Sum the token, segment, and positional embeddings to obtain the complete
        # embedding tensor.
        x = self.token_embeddings(token_ids)
        x = x + self.segment_embeddings(segment_encodings)
        x = x + self.positional_embeddings(position_ids)

        x = self.dropout(x)

        # Pass the embeddings through each transformer block in our model.
        for block in self.transformer_blocks:
            x = block(x, attention_mask)

        return x

    @classmethod
    def from_pretrained(cls, name: str | Path) -> BabyBERT:
        """
        Loads a pretrained `BabyBERT` model from a directory. The directory should
        contain a file named `config.json`, which stores the configuration settings,
        and a file named `pytorch_model.bin`, which contains the model weights.

        Args:
            name: The directory containing the `BabyBERT` model configuration settings
                  and weights.
        Returns:
            An instance of `BabyBERT` with configuration settings and weights loaded
            from the input directory.
        """
        name = resolve_checkpoint_path(name)

        with open(name / "config.json", "r") as config_file:
            config_dict = json.load(config_file)
        config = BabyBERTConfig(**config_dict)

        model = cls(config)

        state_dict = torch.load(name / "pytorch_model.bin", weights_only=True)
        model.load_state_dict(state_dict)

        return model

    def save_pretrained(self, name: str | Path) -> None:
        """
        Saves a pretrained `BabyBERT` model in a directory for later use. Creates two
        new files in the directory: one called `config.json`, which contains
        configuration settings, and another called `pytorch_model.bin`, which contains
        the model weights.

        Args:
            name: The directory in which to save the model's configuration settings and
                  weights.
        """
        name = Path(name)
        name.mkdir(parents=True, exist_ok=True)

        with open(name / "config.json", "w") as config_file:
            json.dump(self.config.__dict__, config_file)

        torch.save(self.state_dict(), name / "pytorch_model.bin")


class BabyBERTForMLM(nn.Module):
    """BabyBERT model with added masked language modeling (MLM) head."""

    def __init__(self, bert: BabyBERT):
        super().__init__()

        self.bert = bert
        self.config = self.bert.config

        # Projection from each BabyBERT contextual embedding to the vocabulary space.
        self.mlm_head = nn.Linear(self.config.hidden_size, self.config.vocab_size)

    def forward(self, token_ids, attention_mask=None, labels=None):
        x = self.bert(token_ids, attention_mask)

        # Obtain the logits for each contextual embedding - we'll only use the logits
        # from the [MASK] tokens.
        mlm_logits = self.mlm_head(x)

        loss = None

        if labels is not None:
            # Flatten the logit and label vectors, rendering them suitable inputs for
            # PyTorch's cross entropy loss function.
            flattened_logits = mlm_logits.view(-1, mlm_logits.size(-1))
            flattened_labels = labels.view(-1)

            # Here's where the important stuff happens - at each masked position in
            # the input sequence, we compare the predicted token probabilities to
            # the true token IDs. If they match up, we have a low loss value; if
            # they're substantially different, our loss will be higher.
            loss = F.cross_entropy(
                flattened_logits,
                flattened_labels,
                ignore_index=self.config.ignore_index,
            )

        return mlm_logits, loss


class BabyBERTForSentimentClassification(nn.Module):
    """BabyBERT model with added sentiment classification head."""

    def __init__(self, bert: BabyBERT):
        super().__init__()

        self.bert = bert
        self.config = self.bert.config

        # Projection from each BabyBERT classification token embedding to a predicted
        # sentiment score (positive or negative)
        self.sentiment_classification_head = nn.Linear(self.config.hidden_size, 2)

    def forward(self, token_ids, attention_mask=None, labels=None):
        x = self.bert(token_ids, attention_mask)

        # Obtain the [CLS] token embedding from each sequence in the batch.
        cls_embeddings = x[:, 0, :]

        # Use our sentiment analysis head to predict a probability for each sentiment
        # class using the [CLS] embedding.
        sentiment_scores = self.sentiment_classification_head(cls_embeddings)

        loss = None

        if labels is not None:
            # Calculate the difference between the class predictions and the ground
            # truth label
            loss = F.cross_entropy(input=sentiment_scores, target=labels)

        return sentiment_scores, loss



import torch

from babybert.model import BabyBERT, BabyBERTConfig


def test_forward_pass():
    config = BabyBERTConfig(
        vocab_size=5,
        hidden_size=10,
        block_size=5,
        n_blocks=1,
        n_heads=10,
        attention_dropout_probability=0.1,
        attention_projection_dropout_probability=0.1,
        mlp_dropout_probability=0.1,
        embedding_dropout_probability=0.1,
    )

    input = torch.asarray([[1, 2, 0, 4, 3], [1, 2, 0, 3, 4]])
    model = BabyBERT(config)
    output = model(input)
    assert output.shape == (2, 5, 10)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable

import torch
from torch.utils.data import DataLoader, Dataset
from tqdm import tqdm

if TYPE_CHECKING:
    import torch.nn as nn


@dataclass
class TrainerConfig:
    """
    Configuration class for a trainer.

    Attributes:
        device: The device on which to perform training.
        learning_rate: The fixed learning rate to use.
        batch_size: The size of each training batch.
        num_workers: The number of subprocess workers to use for loading data.
        collator: The function to use for batching samples.
        num_samples: The number of samples to use during training.
        max_grad_norm: The max norm to which gradient are clipped.
    """

    device: str = "cpu"
    learning_rate: float = 1e-4
    batch_size: int = 64
    num_workers: int = 0
    collator: Callable | None = None
    num_samples: int = 100_000
    max_grad_norm: float = 1


class Trainer:
    """Simple model-agnostic PyTorch trainer class."""

    def __init__(self, model: nn.Module, config: TrainerConfig):
        self.model = model
        self.config = config
        self.device = config.device

    def run(self, data: Dataset):
        """
        Train the BabyBERT model on the provided dataset.

        Args:
            data: The `Dataset` to use for training.
        """
        model = self.model
        config = self.config

        optimizer = torch.optim.Adam(model.parameters(), config.learning_rate)

        loader = DataLoader(
            data,
            batch_size=config.batch_size,
            num_workers=config.num_workers,
            collate_fn=config.collator,
            sampler=torch.utils.data.RandomSampler(
                data, replacement=True, num_samples=config.num_samples
            ),
        )

        model.train()

        pbar = tqdm(
            loader,
            colour="yellow",
            desc="Training",
            unit="samples",
            unit_scale=config.batch_size,
        )

        for batch in pbar:
            batch = [sample.to(self.device) for sample in batch]
            x, masks, y = batch
            _, loss = model(x, mask=masks, labels=y)

            pbar.set_postfix({"loss": f"{loss:.4f}"})

            optimizer.zero_grad()

            loss.backward()
            torch.nn.utils.clip_grad_norm_(model.parameters(), config.max_grad_norm)
            optimizer.step()



from __future__ import annotations

import json
import re
import warnings
from collections import defaultdict
from dataclasses import dataclass, field
from pathlib import Path

import torch

from babybert.utils import resolve_checkpoint_path


@dataclass
class TokenizerConfig:
    """
    Configuration class for a tokenizer.

    Attributes:
        _special_tokens: A list of all special tokens that's down have their own
                         attributes.
        padding_token: The token used for padding sequences to a specified length.
        unknown_token: The token used for representing words or subwords outside of the
                       tokenizer's vocabulary.
        mask_token: The token used for masking sequences during masked language
                    modeling.
        classification_token: The token used for downstream classification tasks.
        target_vocab_size: The vocab size that the tokenizer should attempt to reach.
                           If the corpus used for training is too small, it is possible
                           that the vocab of the tokenizer may not reach this size.
    """

    _special_tokens: list[str] = field(default_factory=lambda: ["[SEP]"])
    padding_token: str = "[PAD]"
    unknown_token: str = "[UNK]"
    mask_token: str = "[MASK]"
    cls_token: str = "[CLS]"
    target_vocab_size: int = 1000

    @property
    def special_tokens(self):
        return [
            *self._special_tokens,
            self.padding_token,
            self.unknown_token,
            self.mask_token,
            self.cls_token,
        ]


class WordPieceTokenizer:
    """Basic implementation of WordPiece tokenizer."""

    def __init__(self, config: TokenizerConfig = None):
        self.config = config or TokenizerConfig()
        self.vocab = []

    @staticmethod
    def _pretokenize(text: str) -> list[str]:
        """
        Perform pretokenization on a text. Carries out the following steps:

        - Converts the text to lowercase.
        - Splits on and removes whitespaces.
        - Splits on and preserves punctuation.

        Args:
            text: The text to pretokenize.
        Returns:
            A list containing each part of the pretokenized text.

        Example:
            >>> WordPieceTokenizer._pretokenize("Hello, world!")
            ["hello", ",", "world", "!"]
        """
        text_lower = text.lower()
        parts = re.split(r"\s|([.!?,;])", text_lower)
        return [part for part in parts if part]

    @staticmethod
    def _score_pair(
        first_token: str, second_token: str, pair_freq: int, token_freqs: dict[str, int]
    ) -> float:
        """
        Computes the WordPiece score of a pair of tokens using the following
        formula:

        score = frequency(pair) / (frequency(first) * frequency(second))

        Args:
            first_token: The first token in the pair.
            second_token: The second token in the pair.
            pair_freq: The frequency at which the pair occurs in the corpus.
            token_freqs: The dictionary containing token to frequency mappings.
        Returns:
            The WordPiece pair score, which is the pair's frequency normalized by the
            product of the individual token frequencies.
        """
        return pair_freq / (token_freqs[first_token] * token_freqs[second_token])

    @staticmethod
    def _compute_pair_scores(tokenized: list[str]) -> dict[tuple[str, str], float]:
        """
        Compute the WordPiece score for each consecutive pair of tokens in the corpus.

        Args:
            tokenized: The list of tokens.
        Returns:
            A dictionary, containing a mapping between consecutive token pairs and
            their scores.
        """
        token_freqs = defaultdict(int)
        pair_freqs = defaultdict(int)

        for tokens in tokenized:
            for token in tokens:
                token_freqs[token] += 1
            for pair in zip(tokens, tokens[1:]):
                pair_freqs[pair] += 1

        scores = {
            pair: WordPieceTokenizer._score_pair(*pair, freq, token_freqs)
            for pair, freq in pair_freqs.items()
        }
        return scores

    @staticmethod
    def _merge_tokens(
        pair: tuple[str, str], replacement: str, words: list[list[str]]
    ) -> dict[str, list[str]]:
        """
        Finds instances of the specified pair of consecutive tokens and replaces them
        with a new token.

        Args:
            pair: The pair of consecutive tokens to find and replace.
            replacement: The token with which to replace the pair of tokens.
            words: The words in which to replace the tokens.
        Returns:
            The list of words, with all instances of the target pair merged.

        Example:
            >>> WordPieceTokenizer._merge_tokens(
                ("a", "##b"),
                "ab",
                [["a", "##b", "##s"], ["a", "##r", "##c"], ["a", "##b", "##late"]]
            )
            [["ab", "##s"], ["a", "##r", "##c"], ["ab", "##late"]]
        """
        for tokens in words:
            i = 0
            while i < len(tokens) - 1:
                current = (tokens[i], tokens[i + 1])
                if current == pair:
                    tokens.pop(i + 1)
                    tokens[i] = replacement
                i += 1
        return words

    @staticmethod
    def _make_replacement(a: str, b: str) -> str:
        """
        Make a merged token from a pair of tokens.

        Args:
            a: The first token to merge.
            b: The second token to merge.
        Returns:
            The token created by merging.

        Example:
            >>> WordPieceTokenizer._make_replacement("h", "##i")
            "hi"
        """
        return a + (b[2:] if b.startswith("##") else b)

    @staticmethod
    def _get_character_tokens(word: str):
        """
        Split a word into its component character tokens, using the WordPiece style.
        This includes a "##" prefix before each non-leading character.

        Args:
            word: The word to split.
        Returns:
            A list of character tokens made from splitting the input.

        Examples:
            >>> WordTokenizer._get_character_tokens("banana")
            ["b", "##a", "##n", "##a", "##n", "##a"]
        """
        return [
            character if i == 0 else f"##{character}"
            for i, character in enumerate(word)
        ]

    def train(self, corpus: list[str]) -> None:
        """
        Trains a tokenizer using the WordPiece protocol. The training comprises the
        following steps:

        1. Pretokenizes the corpus, performing basic operations such as splitting on
        whitespace and punctuation.
        2. Converts each word in the corpus to character tokens. These character tokens
        make up the initial state of our vocabulary.
        3. Computes a score for each pair of consecutive tokens in the corpus, denoting
        how likely they are to occur next to each other.
        4. Selects the pair with the highest score to merge, replacing instances of the
        component tokens with the merged token and adding the merged token to the
        vocabulary.
        5. Repeats steps 3 and 4 until the desired vocabulary size is reached, or no
        more merges can be made (meaning each entire word in the corpus is in the
        vocabulary.)

        Args:
            corpus: The list of texts to use for training. Should contain a
            diverse set of words and punctuation marks.
        """
        pretokenized_words = [
            word for text in corpus for word in WordPieceTokenizer._pretokenize(text)
        ]

        word_tokens = [
            WordPieceTokenizer._get_character_tokens(word)
            for word in pretokenized_words
        ]

        self.vocab = list(set(token for word in word_tokens for token in word))
        self.vocab.extend(self.config.special_tokens)

        while len(self.vocab) < self.config.target_vocab_size:
            scores = self._compute_pair_scores(word_tokens)

            if not scores:
                warnings.warn(
                    "No more pairs left to merge, "
                    f"stopping training with {len(self.vocab)} tokens in vocabulary."
                )
                break

            merge_pair = max(scores, key=scores.get)
            replacement = self._make_replacement(*merge_pair)
            self.vocab.append(replacement)
            word_tokens = self._merge_tokens(merge_pair, replacement, word_tokens)

    @staticmethod
    def _make_all_prefix_substrings(string: str) -> list[str]:
        """
        Generates all prefix substrings from a given string.
        Utility function for `_encode_word`.

        Args:
            string: The string for which to generate the substrings.
        Returns:
            The list of all prefix substrings possible.

        Examples:
            >>> WordPieceTokenizer._make_all_prefix_substrings("garage")
            ["garage", "garag", "gara", "gar", "ga", "g", ""]
        """
        return [string[:-i] if i != 0 else string for i in range(len(string) + 1)]

    def _encode_word(self, word: str) -> list[str]:
        """
        Encodes a word using a longest-first matching strategy. Uses tokens from the
        vocabulary to encode the word.

        Args:
            word: The word to encode.
        Returns:
            The word encoded as tokens from the vocabulary.
        """
        vocab = self.vocab
        tokens = []

        while word:
            for substring in WordPieceTokenizer._make_all_prefix_substrings(word):
                if substring in vocab:
                    tokens.append(substring)
                    if len(substring) == len(word):
                        return tokens
                    else:
                        word = f"##{word[len(substring) :]}"
                        break
                elif not substring:
                    return [self.config.unknown_token]

    def tokenize(self, text: str) -> list[str]:
        """
        Tokenizes a text, converting it into a set of tokens contained in the
        tokenizer's vocabulary.

        Args:
            text: The text to tokenize.
        Returns:
            A list of tokens generated from encoding the input text.
        """
        pretokenized_words = WordPieceTokenizer._pretokenize(text)
        encoded = [
            token for word in pretokenized_words for token in self._encode_word(word)
        ]
        return [self.config.cls_token] + encoded

    def get_token_index(self, token: str) -> int:
        """
        Gets the index of a token in the tokenizer's vocabulary.

        Args:
            token: The token for which to find the index.
        Returns:
            The token's index in the vocabulary.
        """
        return self.vocab.index(token)

    def encode(
        self,
        text: str,
        padding: bool = False,
        padding_length: int | None = None,
    ) -> list[int]:
        """
        Encodes a text as a list of token IDs.

        Args:
            text: The text to encode.
            padding: `True` if output list should be padded to a certain length,
                     `False` otherwise.
            padding_length: The length to which the output lists should be padded.
        Returns:
            The token IDs of the tokenized text.
        """
        tokens = self.tokenize(text)
        token_ids = [self.get_token_index(token) for token in tokens]
        if padding:
            token_ids += [self.padding_token_id] * (padding_length - len(token_ids))
        return token_ids

    def batch_encode(
        self, texts: list[str], padding: bool = True, padding_length: int | None = None
    ) -> dict[str, torch.Tensor]:
        """
        Encodes a batch of texts as token ID lists.

        Args:
            texts: The list of texts to encode.
            padding: `True` if all the output token ID lists should be padded to a
                     certain length, `False` otherwise.
            padding_length: The length to which the output token ID lists should be
                            padded. If `None`, the output lists will be padded to the
                            length of the longest sequence in the batch.
        Returns:
            A dictionary with two entries: 'token_ids', containing the token IDs for
            each text sequence in the batch, and 'attention_mask', containing the
            attention masks for each sequence in the batch.
        """
        tokens = [self.tokenize(text) for text in texts]

        if padding_length is None:
            padding_length = max(len(t) for t in tokens)

        token_ids = [self.encode(text, padding, padding_length) for text in texts]

        attention_mask = [
            [0 if token_id == self.padding_token_id else 1 for token_id in toks]
            for toks in token_ids
        ]

        return {
            "token_ids": torch.tensor(token_ids),
            "attention_mask": torch.tensor(attention_mask),
        }

    def decode(
        self, token_ids: list[int], ignore_special_tokens: bool = True
    ) -> list[str]:
        """
        Decodes a text from a list of token IDs.

        Args:
            token_ids: The token IDs to decode.
            ignore_special_tokens: `True` if special tokens should be omitted from the
                                   output, `False` otherwise.
        Returns:
            The text decoded from the token IDs.
        """
        return [
            token
            for index in token_ids
            if (token := self.vocab[index]) not in self.config.special_tokens
            or not ignore_special_tokens
        ]

    def batch_decode(
        self, token_ids: list[list[int]], ignore_special_tokens: bool = True
    ) -> list[list[str]]:
        """
        Decodes texts from a batch of token IDs.

        Args:
            token_ids: The batch of token IDs to decode.
            ignore_special_tokens: `True` if special tokens should be omitted from the
                                   the output, `False` otherwise.
        Returns:
            The batch of texts decoded from the token IDs.
        """
        return [self.decode(tokens, ignore_special_tokens) for tokens in token_ids]

    @property
    def vocab_size(self) -> int:
        """The number of tokens in the vocabulary."""
        return len(self.vocab)

    @classmethod
    def from_pretrained(cls, name: str | Path) -> WordPieceTokenizer:
        """
        Load a pretrained tokenizer from a directory.

        Args:
            name: The name of the directory from which to load the model.
        Returns:
            A `WordPieceTokenizer` with a pretrained vocabulary and preset
            configuration.
        """
        name = resolve_checkpoint_path(name)

        with open(name / "tok_config.json", "r") as config_file:
            config_dict = json.load(config_file)
        config = TokenizerConfig(**config_dict)

        tokenizer = cls(config)

        with open(name / "vocab.txt", "r") as vocab_file:
            vocab = [line.strip() for line in vocab_file]

        tokenizer.vocab = vocab

        return tokenizer

    def save_pretrained(self, name: str | Path) -> None:
        """
        Save a pretrained tokenizer to a file.

        Args:
            directory: The directory to which to save the pretrained tokenizer.
        """
        name = Path(name)
        name.mkdir(parents=True, exist_ok=True)

        with open(name / "tok_config.json", "w") as config_file:
            json.dump(self.config.__dict__, config_file)

        with open(name / "vocab.txt", "w") as vocab_file:
            for token in self.vocab:
                vocab_file.write(f"{token}\n")

    @property
    def special_token_ids(self) -> list[int]:
        return [self.get_token_index(token) for token in self.config.special_tokens]

    @property
    def padding_token_id(self) -> int:
        return self.get_token_index(self.config.padding_token)

    @property
    def mask_token_id(self) -> int:
        return self.get_token_index(self.config.mask_token)



from __future__ import annotations

from collections import defaultdict
from pathlib import Path
from typing import TYPE_CHECKING

import torch
from torch.utils.data import Dataset

if TYPE_CHECKING:
    from tokenizer import WordPieceTokenizer


class LanguageModelingDataset(Dataset):
    """Class for storing an LM dataset."""

    def __init__(
        self,
        token_ids: torch.Tensor,
        attention_mask: torch.Tensor,
        labels: list[int] | torch.Tensor | None = None,
    ):
        self.token_ids = token_ids
        self.attention_mask = attention_mask
        self.labels = torch.tensor(labels) if labels else None

    def __len__(self) -> int:
        return len(self.token_ids)

    def __getitem__(
        self, index: int
    ) -> (
        tuple[torch.Tensor, torch.Tensor, torch.Tensor]
        | tuple[torch.Tensor, torch.Tensor]
    ):
        if self.labels is not None:
            return self.token_ids[index], self.attention_mask[index], self.labels[index]
        else:
            return self.token_ids[index], self.attention_mask[index]

    @property
    def seq_length(self) -> int:
        """Returns the length of the sequences in the dataset."""
        return self.token_ids.size(1)

    @classmethod
    def from_dict(cls, data: dict[str, torch.Tensor]) -> LanguageModelingDataset:
        """
        Create a new dataset instance from a dictionary.

        Args:
            data: A dictionary containing entries `token_ids`, `attention_mask`, and
                  optionally `labels`.
        Returns:
            A new dataset object with values populated from the dictionary.
        """
        return cls(
            data.get("token_ids"), data.get("attention_mask"), data.get("labels")
        )


class CollatorForMLM:
    """Data collator for applying masks to batches of tokens."""

    def __init__(
        self,
        tokenizer: WordPieceTokenizer,
        mask_prob: float = 0.1,
        ignore_index: int = -100,
    ):
        self.tokenizer = tokenizer
        self.mask_prob = mask_prob
        self.ignore_index = ignore_index

    def __call__(
        self, batch: list[tuple[torch.Tensor, torch.Tensor]]
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Perform collation and masking on the input batch.

        Args:
            batch: The batch of tokens to collate and mask.
        Returns:
            The masked token IDs, the attention mask, and the labels for masked tokens.
        """
        token_ids, attention_mask = zip(*batch)

        batched_token_ids = torch.stack(token_ids)
        batched_attention_mask = torch.stack(attention_mask)

        masked_token_ids, labels = self._mask_tokens(batched_token_ids)

        return masked_token_ids, batched_attention_mask, labels

    def _mask_tokens(
        self, token_ids: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Masks tokens in a batch of token IDs, returning the masked input sequence and
        the labels.

        Args:
            token_ids: The batch of token IDs to mask.
        Returns:
            The tensor containing the batch of token IDs after performing masking and
            the tensor containing the labels (-100 for unmasked tokens).
        """
        # This is where we generate the MLM mask. We begin by generating a random value
        # between 0 and 1 for each token ID in the batch. Then, each random value below
        # our masking probability value becomes a masked token (a `True` in the mask),
        # and all other remain unmasked (a `False` in the mask).
        probs = torch.rand_like(token_ids, dtype=torch.float32)
        mlm_mask = probs < self.mask_prob

        # We also want to make sure that we don't accidentally mask a special token;
        # our model doesn't need to know how to predict a padding token! To do this,
        # we create another mask for special tokens.
        special_tokens_mask = ~torch.isin(
            token_ids,
            torch.tensor(self.tokenizer.special_token_ids, device=token_ids.device),
        )

        # We combine our MLM and special tokens masks together to form our final mask.
        mask = mlm_mask & special_tokens_mask

        # Finally, we perform the masking: all tokens selected for masking are replaced
        # with "[MASK]".
        masked_token_ids = token_ids.masked_fill(mask, self.tokenizer.mask_token_id)

        # We only want labels for the masked tokens, as that's what we're trying to
        # predict during training. All other tokens are set to the ignore index in the
        # label tensor (and consequently ignored by our loss function.)
        labels = token_ids.masked_fill(~mask, self.ignore_index)
        return masked_token_ids, labels


def load_corpus(path: str | Path) -> list[str]:
    """
    Loads a corpus from a `.txt` file. Assumes that each sentence is separated by a
    newline.

    Args:
        path: The path to the `.txt` file containing the corpus.
    Returns:
        A `list[str]` containing each sentence from the corpus.
    """
    path = Path(path)
    return path.read_text().split("\n")


def load_dataset(path: str | Path) -> dict[str, list[str] | list[int]]:
    """
    Loads a dataset from a `.txt` file.

    Args:
        path: The path to the `.txt` file containing the dataset. The contents of
              the file should be of the following form:

              ```
              <text 1>
              <integer label for text 1>
              <text 2>
              <integer label for text 2>
              ...
              ```
    Returns:
        A new dataset object with values populated from the text file.
    """
    lines = load_corpus(path)

    if len(lines) % 2 != 0:
        raise ValueError(
            f"The file '{path}' contains an invalid dataset. Make sure that your"
            "file contains alternating lines of texts and labels."
        )

    data_dict = defaultdict(list)
    for text, label in zip(lines[::2], lines[1::2]):
        data_dict["text"].append(text)
        data_dict["label"].append(int(label))
    return dict(data_dict)



from __future__ import annotations

import importlib.resources
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from importlib.abc import Traversable


def resolve_checkpoint_path(name: str | Path) -> Path | Traversable:
    """
    Resolves a checkpoint name as either a local directory or a directory in the
    library.

    Prioritizes resolving the name with a local directory. If no local directory
    called `name` is discovered, then the function falls back to searching for a
    pretrained checkpoint in the library's `checkpoints/` directory.

    Args:
        name: The name to resolve. Can either be the name of a local directory of the
              name of a pretrained checkpoint included in the library.
    Returns:
        The resolved checkpoint path.
    """
    # If `name` is already a `Path`, simply return it.
    if isinstance(name, Path):
        return name

    # If `name` is the name of a local directory, return the path to the local
    # directory.
    if (local_dir := Path(name)).is_dir():
        return local_dir

    # Otherwise, search for the checkpoint in the library's `checkpoints/`
    # directory.
    checkpoints_dir = importlib.resources.files("babybert.checkpoints")

    if (checkpoint_dir := checkpoints_dir / name).is_dir():
        return checkpoint_dir

    available_checkpoints = ", ".join(dir.name for dir in checkpoints_dir.iterdir())

    raise FileNotFoundError(
        f"No checkpoint '{name}' was found locally or in the library's pretrained "
        f"checkpoints directory. Available checkpoints: {available_checkpoints}"
    )
